/* 
 * tclJava.c --
 *
 *	This file implements the native methods for the tcl.Interp class.
 *
 * Copyright (c) 1995 by Sun Microsystems, Inc.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * SCCS: @(#) tclJava.c 1.2 95/12/19 14:34:48
 */
#include <tcl.h>

#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <fcntl.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <oobj.h>
#include <interpreter.h>
#include <common_exceptions.h>
#include <java_lang_String.h>
#include <javaString.h>
#include <typecodes.h>

#include "tcl_classes.h"

typedef struct CommandInfo {
    Htcl_Command *cmd;
    Htcl_Interp *interp;
} CommandInfo;

static ClassClass *value_class;

/*
 * This is the callback that all TCL commands written in Java actually call.
 * Since all callback objects have the invoke method, this is not a problem.
 */
static int
tcl_invokeCmd(ClientData clientData, Tcl_Interp *interp,
	 int argc, char *argv[])
{
    CommandInfo *info = (CommandInfo *) clientData;
    HArrayOfObject *java_args;
    struct execenv *ee = EE();
    int i, n, ret;
    struct Htcl_Value *rval;
    
    /*
     * Collect all the tcl arguments, and put them into a new Java
     * array.
     * 
     * God bless garbage collection.
     */
    java_args = (HArrayOfObject *) ArrayAlloc(T_CLASS, argc);
    for (i = 0; i < argc; i++) {
	unhand(java_args)->body[i] = (HObject *)execute_java_constructor(ee,
		0, value_class, "(Ljava/lang/String;)",
		makeJavaString(argv[i], strlen(argv[i])));
    }

    /*
     * invoke the java command.
     */
    rval = (struct Htcl_Value *) execute_java_dynamic_method(ee,
            (HObject *)info->cmd, "invoke", 
	    "(Ltcl/Interp;[Ltcl/Value;)Ltcl/Value;", info->interp, java_args);
    if (exceptionOccurred(ee)) {
	/* If the exception is a subclass of Tcl_ResultCodeException, 
	 * set the result to an appropriate value and return the
	 * appropriate code.  (We have to be careful about what we set
	 * the result to.)  Otherwise, we return TCL_ERROR and indicate
	 * that some other Java exception occurred.
	 */
	if (is_instance_of(ee->exception.exc,
	        FindClass(ee, "tcl/Tcl_UserException", FALSE), ee)) {
	    /* !!! Not sure what to do with descrip here.  My guess is
	     *     that it should be placed in the result string iff
	     *     the ERR_ALREADY_LOGGED flag is not set.  However, it's
	     *     also probably true that if that flag is set, the
	     *     interpreter will ignore changes to the result until
	     *     the error is caught, so it might not matter what we
	     *     do.
	     */
            Classtcl_Tcl_0005fUserException *exc = 
		    (Classtcl_Tcl_0005fUserException *)
		    unhand(ee->exception.exc);
	    char *descrip = makeCString(
	            (struct Hjava_lang_String *)
	            execute_java_dynamic_method(ee, (HObject *) exc->descrip,
			    "toString", "()Ljava/lang/String;"));
	    Tcl_SetResult(interp, descrip, TCL_VOLATILE);
	    return (int) execute_java_dynamic_method(ee,
	            (HObject *) ee->exception.exc, "getCode", "()I");
	} else if (is_instance_of(ee->exception.exc,
		FindClass(ee, "tcl/Tcl_ErrorException", FALSE), ee)) {
	    /* !!! See comment above */
            Classtcl_Tcl_0005fErrorException *exc = 
		    (Classtcl_Tcl_0005fErrorException *)
		    unhand(ee->exception.exc);
	    char *descrip = makeCString(
	            (struct Hjava_lang_String *)
	            execute_java_dynamic_method(ee, (HObject *) exc->descrip,
			    "toString", "()Ljava/lang/String;"));
	    Tcl_SetResult(interp, descrip, TCL_VOLATILE);
	    return TCL_ERROR;
	} else if (is_instance_of(ee->exception.exc,
		FindClass(ee, "tcl/Tcl_BreakException", FALSE), ee)) {
	    return TCL_BREAK;
	} else if (is_instance_of(ee->exception.exc,
		FindClass(ee, "tcl/Tcl_ContinueException", FALSE), ee)) {
	    return TCL_CONTINUE;
	} else if (is_instance_of(ee->exception.exc,
		FindClass(ee, "tcl/Tcl_ReturnException", FALSE), ee)) {
            Classtcl_Tcl_0005fReturnException *exc = 
		    (Classtcl_Tcl_0005fReturnException *)
		    unhand(ee->exception.exc);
	    char *result = makeCString(
	            (struct Hjava_lang_String *)
	            execute_java_dynamic_method(ee, (HObject *) exc->result,
			    "toString", "()Ljava/lang/String;"));
	    Tcl_SetResult(interp, result, TCL_VOLATILE);
	    return TCL_RETURN;
	} else {
	    /* Execute the exception's "toString" method to produce a
	     * string representation that we'll append to the result.
	     */
	    char *excstring = makeCString(
	            (struct Hjava_lang_String *)
	            execute_java_dynamic_method(ee,
			    (HObject *) ee->exception.exc,
			    "toString",
			    "()Ljava/lang/String;"));
	    
	    Tcl_SetResult(interp, "Java exception occurred: ", TCL_STATIC);
	    Tcl_AppendResult(interp, excstring, NULL);
	    return TCL_ERROR;
	}
    } else {
	Tcl_SetResult(interp, makeCString(unhand(rval)->string), TCL_VOLATILE);
	ret = TCL_OK;
    }
    
    return ret;
}

/*
 * Oops! A TCL command was deleted.  Let the java object know.
 */
static void
tcl_deleteCmd(ClientData clientData)
{
    CommandInfo *info = (CommandInfo *) clientData;
    execute_java_dynamic_method(0, (HObject *)info->cmd, "deleted",
	    "(Ltcl/Interp;)V", info->interp);
    execute_java_dynamic_method(0, (HObject *)unhand(info->interp)->commands,
	    "removeElement", "(Ljava/lang/Object;)Z", info->cmd);
    ckfree(info);
}    

/*
 * Evaluate a String, returning the result or throwing an exception
 * as necessary.
 */
struct Htcl_Result *
tcl_Interp_native_eval(struct Htcl_Interp *interpH,
		       struct Hjava_lang_String *argH,
		       /* boolean */ long allowExceptions)
{
    int rc;
    struct Htcl_Result *retval;
    Tcl_Interp *interp = (Tcl_Interp *) unhand(interpH)->interp;

    if (allowExceptions) {
	Tcl_AllowExceptions(interp);
    }
    
    rc = Tcl_Eval(interp, makeCString(argH));
    /* Take the return code and result string and pass them back to Java. */
    retval = (struct Htcl_Result *) execute_java_constructor(EE(),
            "tcl/Result", NULL, "(ILjava/lang/String;)",
	    rc, makeJavaString(interp->result, strlen(interp->result)));
    return retval;
}

/*
 * Create the command in TCL-space, making the function defined
 * above "tcl_deleteCmd" the callback command.
 */
void
tcl_Interp_createCommand(Htcl_Interp *interpH, HString *nameH,
	Htcl_Command *cmdH)
{
    Tcl_Interp *interp = (Tcl_Interp *) unhand(interpH)->interp;
    CommandInfo *info = (CommandInfo *) ckalloc(sizeof(CommandInfo));
    
    info->cmd = cmdH;
    info->interp = interpH;

    execute_java_dynamic_method(0, (HObject *)unhand(interpH)->commands,
	    "addElement", "(Ljava/lang/Object;)V", cmdH);

    Tcl_CreateCommand(interp, makeCString(nameH), tcl_invokeCmd,
	    (ClientData) info, tcl_deleteCmd);
}

#ifdef TCLEXECFIX
/*
 * tcl_execcmd
 *
 * This is a grody little hack that will surely only work under unix.
 * it's meant to get around the exec problems that vfork and then
 * exec causes Tcl when executing under Java.
 */
static int
tcl_execcmd(ClientData clientData, Tcl_Interp *interp,
	 int argc, char *argv[])
{
    char buffer[1023];
    char result[1023];
    int i;
    FILE *fp;

    sprintf(result,"%s",argv[1]);
    for(i=2;i<argc;i++) {
        sprintf(buffer, " %s",argv[i]);
        strcat(result,buffer);
    }
    if (!((fp=popen(result,"r")) == -1)) {
        Tcl_ResetResult(interp);
        while(fgets(buffer,1023,fp) != NULL)
            Tcl_AppendResult(interp, buffer, (char *)NULL);
        return pclose(fp);
    } else {
        Tcl_ResetResult(interp);
        sprintf(buffer,"Couldn't exec command '%s'.\n",result);
        Tcl_AppendResult(interp, buffer, (char *) NULL);
        return TCL_ERROR;
    }
}

static int
tcl_deleteexeccmd(ClientData clientData, Tcl_Interp *interp,
	 int argc, char *argv[])
{
    interp = (Tcl_Interp *) clientData;
    Tcl_Eval(interp, "if {[info exists exec.orig]} {rename exec {};rename exec.orig}");
}

#endif {TCLEXECFIX}
/*
 * Initialize TK.
 */
void
tcl_Interp_create(Htcl_Interp *tclH)
{
    Classtcl_Interp *tcl = unhand(tclH);

    /*
     * Initialize static variables, if necessary.
     */

    if (value_class == NULL) {
	value_class = FindClass(0, "tcl/Value", TRUE);
    }

    tcl->interp = (int) Tcl_CreateInterp();
    if (tcl->interp == 0) {
	SignalError(0, JAVAPKG "UnsatisfiedLinkException", "TCL");
	return;
    }

    Tcl_Init((Tcl_Interp *)(tcl->interp));

#ifdef TCLEXECFIX
    Tcl_Eval((Tcl_Interp *)(tcl->interp), "rename exec exec.orig");
    Tcl_CreateCommand((Tcl_Interp *)(tcl->interp), "exec", tcl_execcmd,
	    (Tcl_Interp *)(tcl->interp), tcl_deleteexeccmd);
#endif TCLEXECFIX
    return;
}

/*
 * Do one event, passing appropriate flags as necessary.
 */
long
tcl_Interp_doOneEvent(struct Htcl_Interp *interpH, long flags)
{
#ifdef CREATEMAINWINDOWHACK
    return Tk_DoOneEvent(flags);
#else
    return Tcl_DoOneEvent(flags);
#endif
}

